#include "edimcoder.h"

// Gives back pointer starting at next non-whitespace character that appears after start
char *skipWhitespace(char *start, char *endBound, bool backwards) {
    char *current = start;
	if (current >= endBound) return current;

    if (backwards) {
        do {
            if (current <= endBound) break;
            --current;
        } while (*current == ' ' || *current == '\t' || *current == '\n' || *current == '\r');
    } else {
        while (*current == ' ' || *current == '\t' || *current == '\n' || *current == '\r') {
            ++current;
            if (current >= endBound) break;
        }
    }

    return current;
}

// Will give back a pointer just after the word skipped.
char *skipWord(char *start, char *endBound, bool includeNumbers, bool includeSymbols, bool backwards) {
    char *current = start;
    while ((*current >= 'A' && *current <= 'Z') || (*current >= 'a' && *current <= 'z')
           || (includeNumbers && *current >= '0' && *current <= '9')
           || (includeNumbers && *current == '-') // Note: Technically this will also match numbers that have '-' in the middle and at end, but whatever TODO: Fix this
           || (includeSymbols && *current >= '!' && *current <= '/')
           || (includeSymbols && *current >= ':' && *current <= '@')
           || (includeSymbols && *current >= '[' && *current <= '`')
           || (includeSymbols && *current >= '{' && *current <= '~')) {
        if (backwards && current <= endBound) break;

        if (!backwards) ++current;
        else --current;

        if (!backwards && current >= endBound) break;
    }
    return current;
}

// TODO
char *skipText() {
    return NULL;
}

// TODO
char *skipSymbols() {
    return NULL;
}

char *skipNumbers(char *start, char *endBound) {
    char *current = start;
	if (current >= endBound) return current;
	
    if (*current == '-') ++current; // Note: Not doing '+' here
	if (current >= endBound) return current;
	
    while (*current >= '0' && *current <= '9') {
        ++current;
        if (current >= endBound) break;
    }
    return current;
}

// Specific to Edim Command Language
char *skipLineNumber(char *start, char *endBound) {
    char *current = start;
	if (current >= endBound) return current;
    if (*current == '\'') {
        ++current;
        current = skipWord(current, endBound, true, false, false);
    } else {
        // If symbol
        if ((*current >= '!' && *current <= '/')
            || (*current >= ':' && *current <= '@')
            || (*current >= '[' && *current <= '`')
            || (*current >= '{' && *current <= '~')) {
            //++current;
            if (*current == '.' || *current == '$') // Hacky, TODO
                ++current;
        } else {
            current = skipNumbers(current, endBound);
        }
    }
    return current;
}

// Returns '0' if number could not be parsed
long parseLineNumber(Buffer *buffer, char *start, char *endBound) {
    char *current = start;
	
	if (current >= endBound) return 0;
	
    current = skipWhitespace(current, endBound, false);
    pString lineNumber;
    lineNumber.start = current;
    current = skipLineNumber(current, endBound);
    if (current == lineNumber.start) return 0; // No number found
    lineNumber.end = current;
    
    // Special Line Number Symbols
    if (lineNumber.end - lineNumber.start == 1) {
        switch (lineNumber.start[0]) {
            case '0':
            if (buf_len(buffer->lines) == 0)
                return 0;
            else return 1;
            case '1':
            return 1;
            case '-':
            return 0;
            case '$':
            return buf_len(buffer->lines);
            case '.':
            return buffer->currentLine;
        }
    }
    
    // Turn number string into long integer
    int result = 0;
    bool negative = false;
    current = lineNumber.start;
    if (*current == '-') {
        current++;
		if (current >= endBound) return 0;
		
        negative = true;
    }
    while (*current >= '0' && *current <= '9') {
        result *= 10;
        result += *current - '0';
        ++current;
        if (current == lineNumber.end + 1) break;
		if (current >= endBound) break;
    }
    
    if (negative && result > 0) {
        result *= -1;
    }
    
    return result;
}

/*int parseLineNumber() {

}*/

lineRange parseBookmark(Buffer *buffer, char *start, char *end) {
	char *name_start = start + 1;
	char *current = start;
	
	if (current > end) {
		lineRange line_range = { buffer->currentLine, buffer->currentLine };
		return line_range;
	}
	
	current = skipBookmark(current, end);
	char *name_end = current;
	
	lineRange line_range;
	line_range.start = buffer->currentLine;
	line_range.end = buffer->currentLine;
	
	if (name_end == name_start || name_end == start) { // If no bookmark parsed (or just '#' given)
//		printf("No bookmark or only '#' given\n");
		return line_range;
	}
	
	pString bookmarkName;
	bookmarkName.start = name_start;
	bookmarkName.end = name_end - 1;
	int bookmarkName_length = bookmarkName.end - bookmarkName.start + 1;
	
	if (bookmarkName_length == 1) {
		switch(*bookmarkName.start) {
			case 'p':
			{
			} break;
			
			case 'f':
			{
				line_range.start = 1;
				line_range.end = buf_len(buffer->lines);
			} break;
			
			case 'c':
			{
				line_range.start = buffer->currentLine;
				line_range.end = buffer->currentLine;
			} break;
			
			case 's':
			{
				for (int i = 0; i < buf_len(buffer->bookmarks); i++) {
					if (buffer->currentLine >= buffer->bookmarks[i].range.start && buffer->currentLine <= buffer->bookmarks[i].range.end) {
						line_range.start = buffer->bookmarks[i].range.start;
						line_range.end = buffer->bookmarks[i].range.end;
						break;
					}
				}
			} break;
			
			default:
			{
				Bookmark *result_bookmark;
				bool found = get_bookmark(buffer, bookmarkName, &result_bookmark);
				
				if (found) {
					line_range.start = result_bookmark->range.start;
					line_range.end = result_bookmark->range.end;
				}
			} break;
		}
	} else {
		Bookmark *result_bookmark;
		bool found = get_bookmark(buffer, bookmarkName, &result_bookmark);
//		printf("Result Bookmark Range: %d:%d\n", result_bookmark->range.start, result_bookmark->range.end);
		
		if (found) {
			line_range.start = result_bookmark->range.start;
			line_range.end = result_bookmark->range.end;
		}
	}
	
	return line_range;
}

// Allows A-Za-z0-9 and _ (underscore)
char *skipBookmark(char *start, char *end) {
	char *current = start;
	if (current > end) return current;
	
	if (*current == '#') current++;
	else return current;
	
	if (current > end) return current;
	
	while ((*current >= 'A' && *current <= 'Z') || (*current >= 'a' && *current <= 'z') || (*current >= '0' && *current <= '9') || (*current == '_')) {
		++current;
		
		if (current > end) break;
	}
	
	return current;
}

char *parseLineRange(Buffer *buffer, char *start, char *end, lineRange *line_range) {
	char *current = skipWhitespace(start, end, false);
	int line_start = (int) parseLineNumber(buffer, current, end);
	current = skipLineNumber(current, end);
	
	line_range->start = line_start;
	line_range->end = line_start;
	
	if (current == start) { // If there wasn't a line number passed in
		line_range->start = buffer->currentLine;
		line_range->end = buffer->currentLine;
	
		if (*current == '#') { // Check for bookmark
			lineRange bookmark_range = parseBookmark(buffer, current, end);
			current = skipBookmark(current, end);
			
			line_range->start = bookmark_range.start;
			line_range->end = bookmark_range.end;
		}
	}
	
	int line_range_length = line_range->end - line_range->start;
	
	current = skipWhitespace(current, end, false);
	
	// Relative end, denoted as '+: number' means that the end should be the start plus the number
	// given after '+:'. The end is relative to the start.
	bool relativeEnd = false;
	
	// Check for relative line
	if (*current == '+') { // Addition first
		current++;
		if (current > end) return current;
		current = skipWhitespace(current, end, false);
		
		char *before = current;
		long number = parseLineNumber(buffer, current, end);
		current = skipLineNumber(current, end);
		
		if (current == before) { // If there was no number provided
			relativeEnd = true;
		} else {
			line_range->start = line_range->start + number;
			line_range->end = line_range->end + number;
		}
	} else if (*current == '-') { // Subtraction
		current++;
		if (current > end) return current;
		current = skipWhitespace(current, end, false);
		
		char *before = current;
		long number = parseLineNumber(buffer, current, end);
		current = skipLineNumber(current, end);
		
		if (current == before) { // If there was no number provided
			relativeEnd = true;
		} else {
			line_range->start = line_range->start - number;
			line_range->end = line_range->end - number;
		}
	}
	
	line_range_length = line_range->end - line_range->start;
	current = skipWhitespace(current, end, false);
	
	if (*current == ':') { // Given an end line for the range
		current++;
		if (current > end) return current; // TODO
		current = skipWhitespace(current, end, false);
		
		char *start_endLine = current;
		int line_end = (int) parseLineNumber(buffer, current, end);
		current = skipLineNumber(current, end);
		
		if (relativeEnd) {
			line_range->end = line_range->start + line_end;
		} else {
			line_range->end = line_end;
		}
		
		if (current == start_endLine) { // If there wasn't a line number passed in as end of line range
			line_range->end = buffer->currentLine;
			
			if (*current == '#') { // Check for bookmark as end of range
				lineRange bookmark_range = parseBookmark(buffer, current, end);
				current = skipBookmark(current, end);
				
				line_range->end = bookmark_range.end;
			}
		}
		
		current = skipWhitespace(current, end, false);
		
		line_range_length = line_range->end - line_range->start;
	
		// Check for relative line
		if (*current == '+') { // Addition first
			current++;
			if (current > end) return current;
			current = skipWhitespace(current, end, false);
			
			long number = parseLineNumber(buffer, current, end);
			current = skipLineNumber(current, end);
			
			line_range->end = line_range->end + number;
		} else if (*current == '-') { // Subtraction
			current++;
			if (current > end) return current;
			current = skipWhitespace(current, end, false);
			
			long number = parseLineNumber(buffer, current, end);
			current = skipLineNumber(current, end);
			
			line_range->end = line_range->end - number;
		}
	
		line_range_length = line_range->end - line_range->start;
	}
	
	return current;
}

/* Gets input, trims leading space, and puts into line char array
as long as it doesn't go over the max.

@line - array of characters to fill up
@max - maximum amount of characters allowed in line
@trimSpace - Whether to trim leading space (1) or not (0)
@return - the length of the line
*/
int parsing_getLine(char *line, int max, int trimSpace) {
    int c;
    int i = 0;
    
    /* Trim whitespace */
    while (trimSpace && ((c = getchar()) == ' ' || c == '\t'))
        ;
    
    if (!trimSpace) c = getchar();
    
    /* If there's nothing left, return */
    if (c == '\n') {
        line[i] = '\0';
        return 1; /* Including \0 */
    }
    
    /* Transfer input characters into line string */
    while (c != EOF && c != '\n' && i < max)
    {
        line[i] = (char) c;
        ++i;
        c = getchar();
    }
    
    /* End of string */
    line[i] = '\0';
    ++i; /* Includes '\0' in the length */
    
    return i;
}

// Gets input, trims leading space, and puts into a char stretchy buffer (dynamic array). Does not add null character at end.
// Returns the length of the buffer.
int parsing_getLine_dynamic(char **chars, int trimSpace) {
    int c;
    
    /* Trim whitespace */
    while (trimSpace && ((c = getchar()) == ' ' || c == '\t'))
        ;
    
    if (!trimSpace) c = getchar();
    
    /* If there's nothing left, return */
    if (c == '\n') {
        return 0;
    }
    
    /* Push input characters onto buffer */
    while (c != EOF && c != '\n') {
        buf_push(*chars, (char) c);
        c = getchar();
    }
    
    return buf_len(*chars);
}

// TODO: Test on macOS and BSD!
// TODO: bool printNewLine
// TODO: Placeholder/Ghost text
// TODO: Autocomplete?
char *getInput(bool *canceled, char *inputBuffer, inputKeyCallback callback, unsigned char indentSize) {
    (*canceled) = false;
    int currentIndex = 0;
    int defaultLength = buf_len(inputBuffer);
    
    if (inputBuffer != NULL && buf_len(inputBuffer) > 0) {
        for (int i = 0; i < buf_len(inputBuffer); i++) {
            if (inputBuffer[i] == '\t')
				for (int i = 0; i < indentSize; i++)
					fputc(' ', stdout);
            else if (inputBuffer[i] == INPUT_ESC)
                colors_printf(COLOR_RED, "$");
            else putchar(inputBuffer[i]);
            ++currentIndex;
        }
    }
    
    char c;
    while ((c = getch()) != INPUT_ENDINPUT) { // Ctrl-Z for Windows, Ctrl-D for Linux
#ifdef _WIN32
		if (c == 0) continue;
        if (c == INPUT_SPECIAL1 || c == INPUT_SPECIAL2)
#else
        if (c == INPUT_SPECIAL1)
#endif
        {
#ifndef _WIN32
            if (getch_nonblocking() == INPUT_SPECIAL2) {
#endif
                int specialkey = getch();

				if (callback != NULL) {
					if (!callback(specialkey, true, &inputBuffer, &currentIndex))
						continue;
				}
				
                if (specialkey == INPUT_LEFT) { // Left arrow
                    if (currentIndex != 0) {
                        if (inputBuffer[currentIndex - 1] == '\t')
							for (int i = 0; i < indentSize; i++)
								fputc('\b', stdout);
                        else putchar('\b');
                        --currentIndex;
                    }
                } else if (specialkey == INPUT_RIGHT) { // Right arrow
                    if (currentIndex < buf_len(inputBuffer)) {
                        if (inputBuffer[currentIndex] == '\t')
							for (int i = 0; i < indentSize; i++)
								fputc(' ', stdout);
                        else if (inputBuffer[currentIndex] == INPUT_ESC)
                            colors_printf(COLOR_RED, "$");
                        else putchar(inputBuffer[currentIndex]);
                        ++currentIndex;
                    }
#ifdef _WIN32
                } else if (specialkey == 83) { // Delete
#else
                } else if (specialkey == INPUT_DELETE1) {
                    int special2 = getch();
                    if (special2 == INPUT_DELETE2) {
#endif
                        if (currentIndex < buf_len(inputBuffer)) {
                            // Move all the characters down one
                            char *source = &(inputBuffer[currentIndex + 1]);
                            char *destination = &(inputBuffer[currentIndex]);
                            int amtToMove = buf_end(inputBuffer) - source;
                            memmove(destination, source, sizeof(char) * amtToMove);
                            buf_pop(inputBuffer);
                            // Print all of the characters that have been moved
                            for (int i = currentIndex; i < buf_len(inputBuffer); i++) {
                                if (inputBuffer[i] == '\t')
                                    fputs("    ", stdout);
                                if (inputBuffer[i] == INPUT_ESC)
                                    colors_printf(COLOR_RED, "$");
                                else putchar(inputBuffer[i]);
                            }
                            // Print spaces where the last character(s) used to be
                            fputs("    \b\b\b\b", stdout);
                            // Go back to where the cursor is
                            for (int i = 0; i < buf_len(inputBuffer) - currentIndex; i++) {
                                if (inputBuffer[buf_len(inputBuffer) - 1 - i] == '\t')
                                    fputs("\b\b\b\b", stdout);
                                else putchar('\b');
                            }
                        }
#ifndef _WIN32
                    }
#endif
                } else if (specialkey == INPUT_END) { // End key
                    for (int i = currentIndex; i < buf_len(inputBuffer); i++) {
                        if (inputBuffer[i] == '\t')
							for (int i = 0; i < indentSize; i++)
								fputc(' ', stdout);
                        else if (inputBuffer[i] == INPUT_ESC)
                            colors_printf(COLOR_RED, "$");
                        else putchar(inputBuffer[i]);
                    }
                    currentIndex = buf_len(inputBuffer);
                } else if (specialkey == INPUT_HOME) { // Home key
                    for (int i = 0; i < currentIndex; i++) {
                        if (inputBuffer[i] == '\t')
							for (int i = 0; i < indentSize; i++)
								fputc('\b', stdout);
                        else putchar('\b');
                    }
                    currentIndex = 0;
                    //} else if (special == 115) { // Ctrl+Left // TODO
                    //} else if (special == 116) { // Ctrl+Right // TODO
                } else {
                    //printf("%d", special); // For debugging
                }
                continue;
#ifndef _WIN32
            } else { // TODO: Add to the location of the cursor
                // If escape key was hit and wasn't a special key
                // (most likely ESC hit by itself)
                // then print $ with next character after, and add
                // to buffer ESC + next character
                colors_printf(COLOR_RED, "$");
                //printf("%c", c);
                buf_push(inputBuffer, 27);
                ++currentIndex;
                //buf_push(inputBuffer, c);
                continue;
            }
#endif
        }
        
        if (callback != NULL) {
            if (!callback(c, false, &inputBuffer, &currentIndex))
                continue;
        }
        
        // ASCII Control Keys
        if (c == INPUT_CTRL_X) {
            //buf_pop_all(inputBuffer);
            fputs("^X", stdout);
            buf_free(inputBuffer);
            inputBuffer = NULL;
            currentIndex = 0;
            (*canceled) = true;
            return NULL;
        } else if (c == INPUT_CTRL_C) {
            exit(0); // TODO
        } else if (c > 0 && c <= 26 && c != 13 && c != 8 && c != 9 && c != 10) { // Capture all other Control Keys (aside from Enter, backspace, and tab)
            continue;
        }
        
        if (c == '\n' || c == '\r') { // Enter
#ifdef _WIN32
            putchar('\r');
#endif
            putchar('\n');
            buf_push(inputBuffer, '\n');
            ++currentIndex;
            break;
        }
        if (c == '\t') { // Tab
            if (currentIndex == buf_len(inputBuffer)) {
				for (int i = 0; i < indentSize; i++)
					fputc(' ', stdout);
                buf_push(inputBuffer, '\t');
            } else {
				for (int i = 0; i < indentSize; i++)
					fputc(' ', stdout);
                buf_add(inputBuffer, 1);
                // Move all characters up one
                char *source = &(inputBuffer[currentIndex]);
                char *destination = &(inputBuffer[currentIndex + 1]);
                int amtToMove = &(inputBuffer[buf_len(inputBuffer) - 1]) - source;
                memmove(destination, source, sizeof(char) * amtToMove);
                // Change the character in the inputBuffer
                inputBuffer[currentIndex] = c;
                // Print the characters
                for (int i = currentIndex + 1; i < buf_len(inputBuffer); i++) {
                    if (inputBuffer[i] == '\t')
						for (int i = 0; i < indentSize; i++)
							fputc(' ', stdout);
                    else if (inputBuffer[i] == INPUT_ESC)
                            colors_printf(COLOR_RED, "$");
                    else putchar(inputBuffer[i]);
                }
                // Move back to where the cursor is
                for (int i = 0; i < buf_len(inputBuffer) - currentIndex - 1; i++) {
                    if (inputBuffer[buf_len(inputBuffer) - 1 - i] == '\t')
						for (int i = 0; i < indentSize; i++)
							fputc('\b', stdout);
                    else putchar('\b');
                }
            }
            
            ++currentIndex;
            continue;
        }
		if (c == INPUT_BACKSPACE || c == 8)
        {
            if (currentIndex > 0) {
                if (currentIndex == buf_len(inputBuffer)) {
                    if (inputBuffer[currentIndex - 1] == '\t')
						for (int i = 0; i < indentSize; i++)
							fputc('\b', stdout);
                    else fputs("\b \b", stdout);
                    buf_pop(inputBuffer);
                    --currentIndex;
                    continue;
                }
                
                if (inputBuffer[currentIndex - 1] == '\t')
					for (int i = 0; i < indentSize; i++)
						fputc('\b', stdout);
                else fputs("\b \b", stdout);
                
                // Move all characters down one
                char *source = &(inputBuffer[currentIndex]);
                char *destination = &(inputBuffer[currentIndex - 1]);
                int amtToMove = buf_end(inputBuffer) - source;
                memmove(destination, source, sizeof(char) * amtToMove);
                
                buf_pop(inputBuffer);
                
                // Print all of the characters again
                for (int i = currentIndex - 1; i < buf_len(inputBuffer); i++) {
                    if (inputBuffer[i] == '\t')
						for (int i = 0; i < indentSize; i++)
							fputc(' ', stdout);
                    else if (inputBuffer[i] == '\n' || inputBuffer[i] == '\r')
                        ; // Do Nothing
                    else if (inputBuffer[i] == INPUT_ESC)
                            colors_printf(COLOR_RED, "$");
                    else putchar(inputBuffer[i]);
                }
                
                // Print spaces where the last character(s) used to be
				for (int i = 0; i < indentSize; i++)
					fputc(' ', stdout);
				for (int i = 0; i < indentSize; i++)
					fputc('\b', stdout);
                
                // Move back to where the cursor is
                for (int i = 0; i <= buf_len(inputBuffer) - currentIndex; i++) {
                    if (inputBuffer[buf_len(inputBuffer) - 1 - i] == '\t')
						for (int i = 0; i < indentSize; i++)
							fputc('\b', stdout);
                    else putchar('\b');
                }
                
                --currentIndex;
            }
            continue;
        }

#ifdef _WIN32
        if (c == INPUT_ESC) {
            colors_printf(COLOR_RED, "$");
            buf_push(inputBuffer, c);
            continue;
        }
#endif
        
        if (currentIndex == buf_len(inputBuffer)) {
            putchar(c);
            buf_push(inputBuffer, c);
        } else {
            putchar(c);
            buf_add(inputBuffer, 1);
            // Move all characters up one
            char *source = &(inputBuffer[currentIndex]);
            char *destination = &(inputBuffer[currentIndex + 1]);
            int amtToMove = &(inputBuffer[buf_len(inputBuffer) - 1]) - source;
            memmove(destination, source, sizeof(char) * amtToMove);
            // Change the character in the inputBuffer
            inputBuffer[currentIndex] = c;
            // Print the characters
            for (int i = currentIndex + 1; i < buf_len(inputBuffer); i++) {
                if (inputBuffer[i] == '\t')
					for (int i = 0; i < indentSize; i++)
						fputc(' ', stdout);
                else if (inputBuffer[i] == INPUT_ESC)
                    colors_printf(COLOR_RED, "$");
                else putchar(inputBuffer[i]);
            }
            // Move back to where the cursor is
            for (int i = 0; i < buf_len(inputBuffer) - currentIndex - 1; i++) {
                if (inputBuffer[buf_len(inputBuffer) - 1 - i] == '\t')
					for (int i = 0; i < indentSize; i++)
						fputc('\b', stdout);
                else putchar('\b');
            }
        }
        ++currentIndex;
    }
    
    // If no input was made, free the buffer and return;
    if (buf_len(inputBuffer) - defaultLength <= 0) {
        if (inputBuffer != NULL)
            buf_free(inputBuffer);
    }
    
    // If there's no new line - which there shouldn't be -
    // then add it.
    if (inputBuffer != NULL && *(buf_end(inputBuffer) - 1) != '\n') {
        buf_add(inputBuffer, '\n');
    }
    
    if (buf_len(inputBuffer) == 0) {
#ifdef _WIN32
        printf("^Z");
#else
        printf("^D");
#endif
    }
    
    return inputBuffer;
}

void getFileTypeExtension(FileType ft, char **ftExt) {
    switch (ft) {
        case FT_TEXT:
        {
            (*ftExt) = malloc(sizeof(char) * 3);
            (*ftExt)[0] = 't';
            (*ftExt)[1] = 'x';
            (*ftExt)[2] = 't';
        } break;
        case FT_MARKDOWN:
        {
            (*ftExt) = malloc(sizeof(char) * 2);
            (*ftExt)[0] = 'm';
            (*ftExt)[1] = 'd';
        } break;
        case FT_C:
        {
            (*ftExt) = malloc(sizeof(char) * 1);
            (*ftExt)[0] = 'c';
        } break;
        case FT_CPP:
        {
            (*ftExt) = malloc(sizeof(char) * 3);
            (*ftExt)[0] = 'c';
            (*ftExt)[1] = 'p';
            (*ftExt)[2] = 'p';
        } break;
        case FT_C_HEADER:
        {
            (*ftExt) = malloc(sizeof(char) * 1);
            (*ftExt)[0] = 'h';
        } break;
        case FT_LUA:
        {
            (*ftExt) = malloc(sizeof(char) * 3);
            (*ftExt)[0] = 'l';
            (*ftExt)[1] = 'u';
            (*ftExt)[2] = 'a';
        } break;
    }
}

void createOutline(void) {
    switch (currentBuffer->fileType) {
        case FT_MARKDOWN:
        {
            createMarkdownOutline();
        } break;
		case FT_C_HEADER:
        case FT_C:
        {
            createCOutline();
        } break;
        case FT_TEXT:
        {
            createTextOutline();
        } break;
        case FT_LUA:
        {
            createLuaOutline();
        } break;
    }
}

// Currently, after every operation that modifies the file, the outline is recreated.
// TODO: Perhaps it would be more efficient to only change the line a node points to when
// that line is moved/changed by using some type of hash table for integer keys.
void recreateOutline(void) {
    switch (currentBuffer->fileType) {
        case FT_MARKDOWN:
        {
            // Pop off all of the current nodes
            if (buf_len(currentBuffer->outline.markdown_nodes) > 0)
                buf_pop_all(currentBuffer->outline.markdown_nodes);
            assert(buf_len(currentBuffer->outline.markdown_nodes) == 0);
            createMarkdownOutline();
        } break;
        case FT_C:
        {
            // Pop off all of the current nodes
            if (buf_len(currentBuffer->outline.c_nodes))
                buf_pop_all(currentBuffer->outline.c_nodes);
            assert(buf_len(currentBuffer->outline.c_nodes) == 0);
            createCOutline();
        } break;
        case FT_TEXT:
        {
            // Pop off all of the current nodes
            if (buf_len(currentBuffer->outline.text_nodes))
                buf_pop_all(currentBuffer->outline.text_nodes);
            assert(buf_len(currentBuffer->outline.text_nodes) == 0);
            createTextOutline();
        } break;
        case FT_LUA:
        {
            if (buf_len(currentBuffer->outline.lua_nodes))
                buf_pop_all(currentBuffer->outline.lua_nodes);
            assert(buf_len(currentBuffer->outline.lua_nodes) == 0);
            createLuaOutline();
        } break;
    }
}

void showOutline(void) {
    switch (currentBuffer->fileType) {
        case FT_MARKDOWN:
        {
            showMarkdownOutline();
        } break;
        case FT_C:
        {
            showCOutline();
        } break;
        case FT_TEXT:
        {
            showTextOutline();
        } break;
        case FT_LUA:
        {
            showLuaOutline();
        } break;
    }
}

void createMarkdownOutline(void) {
    assert(currentBuffer->fileType == FT_MARKDOWN);
    int prevLevel = 0;
    lineRange bookmarkRange;
    bookmarkRange.start = -1;
    
    // Go through each line
    for (int line = 0; line < buf_len(currentBuffer->lines); line++) {
        // If starts with a hash, then it's a heading
        if (currentBuffer->lines[line].chars[0] == '#') {
            int level = 0;
            // Increment level with each successive '#'
            for (int i = 1; i < buf_len(currentBuffer->lines[line].chars); i++) {
                if (currentBuffer->lines[line].chars[i] == '#') {
                    level++;
                } else break;
            }
            
            // create the node and push it
            MarkdownOutlineNode node;
            node.line = &(currentBuffer->lines[line]);
            node.lineNum = line;
            node.level = level;
            
//            buf_push(currentBuffer->outline.markdown_nodes, node);

            // Set previous bookmarks ending position, push it onto the
            // bookmarks array, and set starting position of new bookmark.

            if (bookmarkRange.start != -1) { // If there was a previous bookmark
                bookmarkRange.end = line - 1 + 1;

                if (bookmarkRange.end >= buf_len(currentBuffer->lines)) bookmarkRange.end = buf_len(currentBuffer->lines) - 1;

                pString name;
                name.start = currentBuffer->lines[bookmarkRange.start].chars + prevLevel + 1;
                name.end = currentBuffer->lines[bookmarkRange.start].chars + buf_len(currentBuffer->lines[bookmarkRange.start].chars) - 1;

                // Trim starting and ending whitespace of name
                name.start = skipWhitespace(name.start, name.end, false);
                name.end = skipWhitespace(name.end, name.start, true);

                // Fix start to be 1:end rather than 0:end-1 before adding bookmark
                bookmarkRange.start = bookmarkRange.start + 1;
				bool updated = add_bookmark(currentBuffer, name, bookmarkRange, BT_SECTION, true);
            }

            prevLevel = level;
            bookmarkRange.start = line;
        }
    }

    if (bookmarkRange.start != -1) { // If there was a previous bookmark
        bookmarkRange.end = buf_len(currentBuffer->lines);

        pString name;
        name.start = currentBuffer->lines[bookmarkRange.start].chars + prevLevel + 1;
        name.end = currentBuffer->lines[bookmarkRange.start].chars + buf_len(currentBuffer->lines[bookmarkRange.start].chars) - 1;

        // Fix start to be 1:end rather than 0:end-1 before adding bookmark
        bookmarkRange.start = bookmarkRange.start + 1;
		bool updated = add_bookmark(currentBuffer, name, bookmarkRange, BT_FUNCTION, true);
    }
}

void createCOutline(void) {
//    assert(currentBuffer->fileType == FT_C);

    pString name;
    name.start = NULL;
    lineRange bookmarkRange;
    bookmarkRange.start = -1;

    // Go through each line
    for (int line = 0; line < buf_len(currentBuffer->lines); line++) {
        pString currentName;
        
        char *start = currentBuffer->lines[line].chars;
        char *current = start;
        int lineLength = buf_len(currentBuffer->lines[line].chars);
        char *endBound = buf_end(currentBuffer->lines[line].chars);

        //printLine(line, 0, true);

        current = skipWhitespace(current, endBound, false);


        /*printf("      ");
        for (int i = 0; i < current - start - 1; i++) // Debug
            printf(" ");
        printf("^- %c\n", *current);*/

        // Check that has params start ('(')
        char *paramStart = current;
        bool outerContinue = false;
        while (*paramStart != '(') {
            paramStart++;
            if (paramStart >= endBound) {
                outerContinue = true;
                break;
            }
        }

        if (outerContinue) continue;
        //printf("%c\n", *paramStart);

        /*printf("      ");
        for (int i = 0; i < paramStart - start; i++) // Debug
            printf(" ");
        printf("^- %c\n", *paramStart);*/

        // Potential function name
        currentName.end = paramStart - 1;
        while (*currentName.end == ' ') currentName.end = currentName.end - 1;

        // NOTE: without '+ 1', a space is included (this is how it should be)
        currentName.start = skipWord(currentName.end, start, true, true, true) + 1;
        if (*currentName.start == '*') currentName.start = currentName.start + 1;

        // NOTE: currentName.end itself is included in the string/name
        int currentNameLength = currentName.end - currentName.start + 1;

        /* // For Debug:
        printf("       ");
        for (int i = 0; i < currentName.start - start - 1; i++) // Debug
            printf(" ");
        printf("^");
        for (int i = 0; i < currentName.end - currentName.start - 1; i++)
            printf("-");
        printf("^- %c, %c - %.*s\n", *currentName.start, *currentName.end, currentNameLength, currentName.start);
        */

        // Make sure function name doesn't match any language keywords
        if ((strncmp(currentName.start, "if", MAX(currentNameLength, 2)) == 0)
            || (strncmp(currentName.start, "else", MAX(currentNameLength, 4)) == 0)
            || (strncmp(currentName.start, "switch", MAX(currentNameLength, 6)) == 0)
            || (strncmp(currentName.start, "while", MAX(currentNameLength, 5)) == 0)
            || (strncmp(currentName.start, "do", MAX(currentNameLength, 6)) == 0)
            || (strncmp(currentName.start, "case", MAX(currentNameLength, 4)) == 0)
            || (strncmp(currentName.start, "for", MAX(currentNameLength, 3)) == 0)
            || (strncmp(currentName.start, "&&", 2) == 0)
            || (strncmp(currentName.start, "||", 2) == 0)
            || (strncmp(currentName.start, "&", 2) == 0)
            || (strncmp(currentName.start, "|", 2) == 0)
            || (strncmp(currentName.start, "+", 2) == 0)
            || (strncmp(currentName.start, "-", 2) == 0)
            || (strncmp(currentName.start, "*", 2) == 0)
            || (strncmp(currentName.start, "/", 2) == 0)
            || (strncmp(currentName.start, "^", 2) == 0)
            || (strncmp(currentName.start, "//", 2) == 0)
            || (strncmp(currentName.start, "/*", 2) == 0)) {
            continue;
        }

        // Search for types/keywords that signify the line is *not* a function declaration
        //bool hasProperReturnType = true;
        while (true) {
            char *startWord = current;
            current = skipWord(current, currentName.start, true, true, false);
            char *endWord = current;
            current = skipWhitespace(current, currentName.start, false);
            int wordLength = endWord - startWord;

            if ((strncmp(startWord, "if", MAX(wordLength, 2)) == 0)
                || (strncmp(startWord, "else", MAX(wordLength, 4)) == 0)
                || (strncmp(startWord, "switch", MAX(wordLength, 6)) == 0)
                || (strncmp(startWord, "while", MAX(wordLength, 5)) == 0)
                || (strncmp(startWord, "do", MAX(wordLength, 6)) == 0)
                || (strncmp(startWord, "case", MAX(wordLength, 4)) == 0)
                || (strncmp(startWord, "for", MAX(wordLength, 3)) == 0)
                || (strncmp(startWord, "//", 2) == 0)
                || (strncmp(startWord, "/*", 2) == 0)) {
                outerContinue = true;
                break;
            }
            // If matches with any possible return type, set hasReturnType and break
            /*if ((strncmp(startWord, "int", MAX(wordLength, 3)) == 0)
                || (strncmp(startWord, "bool", MAX(wordLength, 4)) == 0)
                || (strncmp(startWord, "float", MAX(wordLength, 5)) == 0)
                || (strncmp(startWord, "double", MAX(wordLength, 6)) == 0)
                || (strncmp(startWord, "char", MAX(wordLength, 4)) == 0)
                || (strncmp(startWord, "string", MAX(wordLength, 6)) == 0)
                || (strncmp(startWord, "void", MAX(wordLength, 4)) == 0)) {
                hasReturnType = true;
                break;
            }*/

            if (current >= currentName.start - 1) break;
        }

        if (outerContinue) continue;

        // Search for paramEnd
        char *paramEnd = paramStart;
        int level = 0;
        while (true) {
            paramEnd++;

            // Don't count nested parentheses
            if (*paramEnd == '(') ++level;
            if (*paramEnd == ')' && level != 0) --level;

            if (paramEnd >= endBound) {
                outerContinue = true;
                break;
            }

            if (*paramEnd == ')' && level == 0) break;
        }

        if (outerContinue) continue;

        // Check For function code start ('{')
        char *codeStart = paramEnd;
        while (*codeStart != '{') {
            codeStart++;
            if (codeStart >= endBound) {
                outerContinue = true;
                break;
            }
        }

		if (outerContinue) {
			continue; // TODO: Disable function declarations without body for now.
			// Check for ';'
			codeStart = paramEnd;
			outerContinue = false;
			while (*codeStart != ';') {
				codeStart++;
				if (codeStart >= endBound) {
					outerContinue = true;
					break;
				}
			}
			
			if (outerContinue) continue;
			else {
				COutlineNode node;
				node.line = &(currentBuffer->lines[line]);
				node.lineNum = line;
				
//				buf_push(currentBuffer->outline.c_nodes, node);
				
				// Add previous bookmark, add this bookmark, set start of next bookmark
				if (bookmarkRange.start != -1) {
					bookmarkRange.end = line - 1 + 1;
					bookmarkRange.start = bookmarkRange.start + 1;
					add_bookmark(currentBuffer, name, bookmarkRange, BT_FUNCTION, true);
				}
				
				bookmarkRange.start = line;
				bookmarkRange.end = line + 1;
				bookmarkRange.start = bookmarkRange.start + 1;
				add_bookmark(currentBuffer, currentName, bookmarkRange, BT_FUNCTION, true);
				
				bookmarkRange.start = -1;
				continue;
			}
		}

        //printf("|%.*s|\n", currentNameLength, currentName.start);
        COutlineNode node;
        node.line = &(currentBuffer->lines[line]);
        node.lineNum = line;
        
//        buf_push(currentBuffer->outline.c_nodes, node);

        if (bookmarkRange.start != -1) {
            // Add previous bookmark, set start of next bookmark
            bookmarkRange.end = line - 1 + 1;
            // Fix start to be 1:end rather than 0:end-1 before adding bookmark
            bookmarkRange.start = bookmarkRange.start + 1;
			add_bookmark(currentBuffer, name, bookmarkRange, BT_FUNCTION, true);
        }

        bookmarkRange.start = line;
        name = currentName;
    }

    if (bookmarkRange.start != -1) {
        // Add previous bookmark, set start of next bookmark
        bookmarkRange.end = buf_len(currentBuffer->lines);
        // Fix start to be 1:end rather than 0:end-1 before adding bookmark
        bookmarkRange.start = bookmarkRange.start + 1;
		add_bookmark(currentBuffer, name, bookmarkRange, BT_SECTION, true);
    }
}

void createTextOutline(void) {
    assert(currentBuffer->fileType == FT_TEXT);

    lineRange bookmarkRange;
    bookmarkRange.start = -1;
    bool blankBefore = true;
    int headerLine = -1;

    // Go through each line
    for (int line = 0; line < buf_len(currentBuffer->lines); line++) {
        // Search for blank line
        if (buf_len(currentBuffer->lines[line].chars) <= 1) {
            if (blankBefore && headerLine != -1) {
                // Add previous heading to bookmarks (including correct line end of section)
                if (bookmarkRange.start != -1) {
                    bookmarkRange.end = headerLine - 1 + 1;

                    if (bookmarkRange.end > buf_len(currentBuffer->lines)) bookmarkRange.end = buf_len(currentBuffer->lines) - 1;

                    pString name;
                    name.start = currentBuffer->lines[bookmarkRange.start].chars;
                    name.end = currentBuffer->lines[bookmarkRange.start].chars + buf_len(currentBuffer->lines[bookmarkRange.start].chars) - 1;

                    // Fix start to be 1:end rather than 0:end-1 before adding bookmark
                    bookmarkRange.start = bookmarkRange.start + 1;
					bool updated = add_bookmark(currentBuffer, name, bookmarkRange, BT_SECTION, true);
                }

                TextOutlineNode node;
                node.line = &(currentBuffer->lines[headerLine]);
                node.lineNum = headerLine;

//                buf_push(currentBuffer->outline.text_nodes, node);

                bookmarkRange.start = headerLine;

                headerLine = -1;
                blankBefore = false;
            }

            blankBefore = true;
        } else if (blankBefore) {
            if (headerLine != -1 && blankBefore) {
                headerLine = -1;
                blankBefore = false;
            }

            headerLine = line;
        }
    }

    if (bookmarkRange.start != -1) {
        bookmarkRange.end = buf_len(currentBuffer->lines);

        pString name;
        name.start = currentBuffer->lines[bookmarkRange.start].chars;
        name.end = currentBuffer->lines[bookmarkRange.start].chars + buf_len(currentBuffer->lines[bookmarkRange.start].chars) - 1;

        // Fix start to be 1:end rather than 0:end-1 before adding bookmark
        bookmarkRange.start = bookmarkRange.start + 1;
		bool updated = add_bookmark(currentBuffer, name, bookmarkRange, BT_SECTION, true);
    }
}

void createLuaOutline(void) {
    assert(currentBuffer->fileType == FT_LUA);

    pString name;
    name.start = NULL;
    lineRange bookmarkRange;
    bookmarkRange.start = -1;

    for (int line = 0; line < buf_len(currentBuffer->lines); line++) {
        pString currentName;

        char *start = currentBuffer->lines[line].chars;
        char *current = start;
        int lineLength = buf_len(currentBuffer->lines[line].chars);
        char *endBound = buf_end(currentBuffer->lines[line].chars);

        current = skipWhitespace(current, endBound, false);

        // Check for 'function' keyword
        char *functionKeyword = "function";
        bool hasKeyword = true;
        for (int i = 0; i < 8; i++) {
            if (current[i] != functionKeyword[i]) {
                hasKeyword = false;
                break;
            }
        }

        if (!hasKeyword) continue;

        current = current + 8;

        current = skipWhitespace(current, endBound, false);
        currentName.start = current;
        current = skipWord(current, endBound, true, false, false);
        currentName.end = current;

        LuaOutlineNode node;
        node.line = &(currentBuffer->lines[line]);
        node.lineNum = line;

//        buf_push(currentBuffer->outline.lua_nodes, node);

        if (bookmarkRange.start != -1) {
            // Add previous bookmark, set start of next bookmark
            bookmarkRange.end = line - 1 + 1;
            // Fix start to be 1:end rather than 0:end-1 before adding bookmark
            bookmarkRange.start = bookmarkRange.start + 1;
			add_bookmark(currentBuffer, name, bookmarkRange, BT_FUNCTION, true);
        }

        bookmarkRange.start = line;
        name = currentName;
    }

    if (bookmarkRange.start != -1) {
        // Add previous bookmark, set start of next bookmark
        bookmarkRange.end = buf_len(currentBuffer->lines);
        // Fix start to be 1:end rather than 0:end-1 before adding bookmark
        bookmarkRange.start = bookmarkRange.start + 1;
		add_bookmark(currentBuffer, name, bookmarkRange, BT_FUNCTION, true);
    }
}

void showMarkdownOutline(void) {
    assert(currentBuffer->fileType == FT_MARKDOWN);
    
    // Go though each node
	for (int node_i = 0; node_i < buf_len(currentBuffer->bookmarks); node_i++) {
		if (!currentBuffer->bookmarks[node_i].outline) continue;
		int linenum = currentBuffer->bookmarks[node_i].range.start - 1;
		
		printLine(linenum, 0, true);
	}
}

void showCOutline(void) {
    assert(currentBuffer->fileType == FT_C);
    
    // Go through each node
	for (int node_i = 0; node_i < buf_len(currentBuffer->bookmarks); node_i++) {
		if (!currentBuffer->bookmarks[node_i].outline) continue;
		int linenum = currentBuffer->bookmarks[node_i].range.start - 1;
		
		printLine(linenum, 0, true);
	}
}

void showTextOutline(void) {
    assert(currentBuffer->fileType == FT_TEXT);

	for (int node_i = 0; node_i < buf_len(currentBuffer->bookmarks); node_i++) {
		if (!currentBuffer->bookmarks[node_i].outline) continue;
		int linenum = currentBuffer->bookmarks[node_i].range.start - 1;
			
		printLine(linenum, 0, true);
	}
}

void showLuaOutline(void) {
    assert(currentBuffer->fileType == FT_LUA);

	for (int node_i = 0; node_i < buf_len(currentBuffer->bookmarks); node_i++) {
		if (!currentBuffer->bookmarks[node_i].outline) continue;
		int linenum = currentBuffer->bookmarks[node_i].range.start - 1;
		
		printLine(linenum, 0, true);
	}
}

/* === Bookmarks === */

bool get_bookmark(Buffer *buffer, pString name, Bookmark **result_bookmark) {
    // NOTE: name.end itself is included in the string/name
    int name_length = name.end - name.start + 1;
    bool found = false;
    for (int i = 0; i < buf_len(buffer->bookmarks); i++) {
        char *bookmarkName = buffer->bookmarks[i].name;
        //printf("Lengths: %d == %d ; ", buf_len(bookmarkName), name_length);
//        printf("Strings: |%.*s|, |%.*s|\n\n", buf_len(bookmarkName), bookmarkName, name_length, name.start);
        if (buf_len(bookmarkName) == name_length && strncmp(bookmarkName, name.start, MIN(buf_len(bookmarkName), name_length)) == 0) { // strncmp(command.start, "clear", maxChars) == 0
            (*result_bookmark) = &buffer->bookmarks[i];
            found = true;
            break;
        }
    }

    return found;
}

bool add_bookmark(Buffer *buffer, pString name, lineRange range, BookmarkType type, bool outline) {
    // NOTE: name.end itself is included in the string/name
    int name_length = name.end - name.start + 1;

    // Check if bookmark already exists // TODO
    Bookmark *result_bookmark;
    bool found = get_bookmark(buffer, name, &result_bookmark);

    if (found) {
        // If already exists, update the range
        result_bookmark->range.start = range.start;
        result_bookmark->range.end = range.end;
	result_bookmark->outline = outline;
	result_bookmark->type = type;
    } else {
        // Otherwise, create and add the bookmark
        Bookmark bookmark;
	bookmark.outline = outline;
	bookmark.type = type;
        bookmark.name = NULL;
        for (int i = 0; i < name_length; i++) { // TODO: Do this before checking if bookmark already exists
            if (name.start[i] == ' ') continue;
            if (name.start[i] == ':') continue;
            if (name.start[i] == ';') continue;
            if (name.start[i] == '!') continue;
            if (name.start[i] == '#') continue;
            if (name.start[i] == ',') continue;
            if (name.start[i] == '.') continue;
            if (name.start[i] == '\'') continue;
            if (name.start[i] == '"') continue;
            if (name.start[i] == '(') continue;
            if (name.start[i] == ')') continue;
            if (name.start[i] == '{') continue;
            if (name.start[i] == '}') continue;
            if (name.start[i] == '<') continue;
            if (name.start[i] == '>') continue;
            if (name.start[i] == '/') continue;
            if (name.start[i] == '\\') continue;
            if (name.start[i] == '*') continue;
            if (name.start[i] == '$') continue;
            if (name.start[i] == '\n') continue;
            if (name.start[i] == '\r') continue;
            if (name.start[i] == '-') {
                buf_push(bookmark.name, '_');
                continue;
            }
            if (name.start[i] == INPUT_ESC) continue;
            buf_push(bookmark.name, name.start[i]);
        }

        bookmark.range.start = range.start;
        bookmark.range.end = range.end;

        buf_push(buffer->bookmarks, bookmark);
    }

    return found;
}
